#! /usr/bin/python

"""Convert Ishi format into SGF.

This module converts a game in Ishi format and saves the result as SGF file.
It is used by glGo and gGo to silently convert Ishi files when loading them
and by the ishi2sgf standalone converter.

Copyright (c) 2003, 2004 Peter Strempel
"""

__version__ = "$Revision: 1.1.1.1 $"
__author__  =  "Peter Strempel"

import glGo.utils, re, os, __init__
from glGo.__init__ import sgftools_version
from string import split

outSGF = ''

class Node:

    """A node within the game tree."""

    def __init__(self, move):
        """Initialize a new Node"""
        self.move = move
        self.comment = ''
        self.parent = None
        self.brother = None
        self.son = None
        self.marks = []
        self.stones = []

    def toSGF(self):
        """Convert Node to SGF"""
        # Move
        if self.move:
            # Pass
            if self.move[1] == 0 and self.move[2] == 0:
                s = ';%s[]' % self.move[0]
            # Empty node (created with AB/AW)
            elif self.move[1] == -1 and self.move[2] == -1:
                s = ';'
            # Normal move
            else:
                s = ';%s[%s%s]' % (self.move[0], self.move[1], self.move[2])
        else:
            s = ''
        # Stones
        if self.stones:
            for stone in self.stones:
                s += 'A%s[%s%s]' % (stone[1], stone[0][0], stone[0][1])
        # Marks
        if self.marks:
            s += 'LB'
            for m in self.marks:
                s += '[%s%s:%s]' % (m[0][0], m[0][1], m[1])
        # Comment
        if self.comment:
            s += 'C[' + self.comment.strip() + ']'
        return s

def findRank(s, headers):
    """Find the rank part a header string."""
    rex = re.compile(r"[0-9]+([pdk]|-(dan|pro|kyu))")
    so = rex.search(s)
    if so:
        return so.group()
    return None

def parseIshi(toParse):
    """Parse the given string with the Ishi game."""
    headers = '(;GM[1]FF[4]AP[%s:%s]ST[1]\n' % (__init__.packagename, sgftools_version)
    comment_flag = 0
    var_flag = 0
    root = Node(None)
    current = root
    inIshi = split(toParse, '\n')
    boardsize = 19

    moveRex = re.compile(r'([BW]) [0-9]+ (([a-z])([0-9]+)|Pass)')
    markRex = re.compile(r'MARK .*')
    markPosRex = re.compile(r'([a-zA-Z0-9]+)@([a-z])([0-9]+)')
    setupRex = re.compile(r'SETUP( [BW] [a-z][0-9]+)+')
    setupStoneRex = re.compile(r'([BW]) ([a-z])([0-9]+)')

    for line in inIshi:
        # Move
        so = moveRex.search(line)
        if so:
            g = so.groups()
            if g[1] == 'Pass':
                x = y = 0
            else:
                x = g[2]
                if ord(x) >= ord('j'):
                    x = chr(ord(x) - 1)
                y = chr(ord('a') + boardsize - int(g[3]))
            n = Node( ( g[0], x, y ) )
            if var_flag:
                n.parent = current.parent
                current.brother = n
                var_flag = 0
            else:
                current.son = n
                n.parent = current
            current = n

        # Variation start/end
        elif line.find('ENDVAR') != -1:
            while 1:
                try:
                    if current != current.parent.son:
                        current = current.parent.son
                        break
                    current = current.parent
                except AttributeError:
                    # No moves in variation. Try to recover by ignoring it.
                    # Don't know if this would be valid or invalid Ishi format
                    var_flag = 0
                    while 1:
                        if not current.son:
                            break
                        current = current.son
                    break
        elif line.find('VAR') != -1:
            var_flag = 1

        # Headers
        # This must be parsed before comments, as in some games I found the headers
        # are inside a comment (smart format...)
        if line.startswith('BOARDSIZE'):
            boardsize = int(line[10:])
            headers += 'SZ[%d]' % boardsize
        elif line.startswith('KOMI'):
            headers += 'KM[%s]' % line[5:]
        elif line.startswith('WHITE: ') or line.startswith('White: '):
            headers += 'PW[%s]' % line[7:].strip()
            r = findRank(line[7:].strip(), headers)
            if r:
                headers += 'WR[%s]' % r
        elif comment_flag and line.startswith('W: '):
            headers += 'PW[%s]' % line[3:].strip()
            r = findRank(line[3:].strip(), headers)
            if r:
                headers += 'WR[%s]' % r
        elif line.startswith('BLACK: ') or line.startswith('Black: '):
            headers += 'PB[%s]' % line[7:].strip()
            r = findRank(line[7:].strip(), headers)
            if r:
                headers += 'BR[%s]' % r
        elif comment_flag and line.startswith('B: '):
            headers += 'PB[%s]' % line[3:].strip()
            r = findRank(line[3:].strip(), headers)
            if r:
                headers += 'BR[%s]' % r

        # Comment start/end
        # (this 'if' is no typo, see above)
        if line.find('ENDCOM') != -1:
            comment_flag = 0
        elif line.find('COM') != -1:
            comment_flag = 1
        elif line.find('EVENT') != -1:
            comment_flag = 2
        # Comment body
        elif comment_flag == 1:
            c = line.strip()
            if c.find('PRISONER') == -1:
                current.comment += c + '\n'

        # Others
        else:
            # Marks
            so = markRex.search(line)
            if so:
                marks = markPosRex.findall(so.group())
                for m in marks:
                    x = m[1]
                    if ord(x) >= ord('j'):
                        x = chr(ord(x) - 1)
                    y = chr(ord('a') + boardsize - int(m[2]))
                    current.marks.append( ( (x, y), m[0] ) )
                continue

            # Setup
            so = setupRex.search(line)
            if so:
                stones = setupStoneRex.findall(so.group()[6:])
                for s in stones:
                    if var_flag:
                        n = Node( ( g[0], -1, -1 ) )
                        n.parent = current.parent
                        current.brother = n
                        var_flag = 0
                        current = n
                    x = s[1]
                    if ord(x) >= ord('j'):
                        x = chr(ord(x) - 1)
                    y = chr(ord('a') + boardsize - int(s[2]))
                    current.stones.append( ( (x, y), s[0] ) )

    global outSGF
    outSGF = ''
    traverse(root)

    return headers + outSGF

def traverse(n):
    """Traverse the game tree and create the SGF from it."""
    global outSGF

    if n.parent:
        outSGF += '\n('
    while 1:
        outSGF += n.toSGF()

        tmp = n.son
        if tmp and tmp.brother:
            while 1:
                traverse(tmp)
                tmp = tmp.brother
                if not tmp:
                    break
            break
        n = n.son
        if n:
            outSGF += '\n'
        else:
            break
    outSGF += ')'

def saveSGF(sgf, filename):
    """Save the converted SGF text to a file."""
    try:
        file = open(filename, 'w')
    except IOError, (errno, strerr):
        print "Error writing SGF to file '%s': %s:" % (filename, strerr)
        return -1

    file.write(sgf)
    file.close()
    return 0

def convertIshi(inFile, outFile):
    """Convert the given Ishi file into the given SGF file.

    This is the main entry method for this module.

    Keyword arguments:
    inFile  -- Input filename in Ishi format
    outFile -- Output filename in SGF format

    Returns:
     0 -- Conversion successful
    -1 -- Conversion failed

    """
    if not inFile or not os.path.exists(inFile):
        print 'No such file: %s' % inFile
        return -1
    if not outFile:
        print 'No outfile given.'
        return -1

    s = glGo.utils.loadFile(inFile)
    if not s:
        print 'No content.'
        return -1
    sgf = parseIshi(s)
    if not sgf:
        print 'Failed to parse file.'
        return -1
    res = saveSGF(sgf, outFile)
    if res == -1:
        print 'Failed to save SGF to', outFile
        return -1
    print 'Converted %s to %s' % (inFile, outFile)
    return 0
